# -*- coding: utf-8 -*-

from robot.utils import ConnectionCache

class ApplicationCache(ConnectionCache):

    def __init__(self):
        ConnectionCache.__init__(self, no_current_msg='No current application')
        self._closed = set()

    def close(self):
        if self.current:
            application = self.current
            application.quit()
            self.current = self._no_current
            self.current_index = None
            self._closed.add(application)

    def close_all(self):
        for application in self._connections:
            if application not in self._closed:
                application.quit()
        self.empty_cache()
        return self.current


# -*- coding: utf-8 -*-

import cv2
import numpy as np
import matplotlib.colors as mcolors
import io
import base64
import os

from PIL import Image
from robot.libraries.BuiltIn import BuiltIn


class ImageProcessing:
    def __init__(self):
        pass
    #KeyWord

    def detect_and_mask_objects(self, image_path , contour_area = None, hsv_array_lower = None, hsv_array_upper = None):
        """
        ***|    Description     |***
        |   *`Detect And Mask Objects`*   |   ส่ง path รูปภาพ และ สี เพื่อไป mask object ที่เป็นสีดังกล่าวภายในภาพ |
 
        ***|    Example     |***
        | *`Detect And Mask Objects`* | image_path | color = String | contour_area = Int
 
        ***|    Parameters     |***
        -  image_path
        -  color (ณ ปัจจุบัน มีให้ทดสอบเฉพาะ color = green)
        -  contour_area (Minimum area to consider a contour valid | หากสนใจพวกเส้นหรือตัวอักษร min_contour_area = 50-100 
        หรือหากสนใจสิ่งที่ใหญ่กว่าเช่นกรอบหรือกล่อง Text Field ให้ min_contour_area = 3500 - 5000++)
        -  hsv_array_lower ค่า min hsv 
        -  hsv_array_upper ค่า max hsv
 
        *`ทดสอบได้เฉพาะบางหน้าเท่านั้น บางรูปอาจจะไม่สามารถจับ mask ต้องกำหนดค่าเฉพาะ`*
        """
        if contour_area == None or hsv_array_lower == None or hsv_array_upper == None:
            raise ValueError("Error Argument contour_area or hsv_array_lower or hsv_array_upper is None ")
        #read image with cv2.imread
        image = cv2.imread(image_path)
        image_old = cv2.imread(image_path)
        # Convert image to HSV color space
        hsv = cv2.cvtColor(image, cv2.COLOR_BGR2HSV)
        
        #np set hsv lower , upper
        h_lower = hsv_array_lower[0]
        s_lower = hsv_array_lower[1]
        v_lower = hsv_array_lower[2]
        h_upper = hsv_array_upper[0]
        s_upper = hsv_array_upper[1]
        v_upper = hsv_array_upper[2]

        # Define the green color range for the mask
        lower_color = None
        upper_color = None
        if contour_area != None and hsv_array_lower != None and hsv_array_upper != None:
            lower_color = np.array([h_lower, s_lower, v_lower])
            upper_color = np.array([h_upper, s_upper, v_upper])

            # Create the mask
            mask = cv2.inRange(hsv,lower_color,upper_color)

            # Find contours in the binary image
            contours, _ = cv2.findContours(mask, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)

            # Filter contours based on their size
            min_contour_area = contour_area  # Minimum area to consider a contour valid
 
            # Filter contours with an area greater than or equal to min_contour_area | กรอง contour ที่มีพื้นที่มากกว่าหรือเท่ากับ min_contour_area
            filtered_contours = [cnt for cnt in contours if cv2.contourArea(cnt) > min_contour_area]

            # Draw contours on the original image with a thin red line | วาดเส้น ตาม filtered_contours ที่มี
            draw_contour_image = cv2.drawContours(image, filtered_contours, -1, (255, 0, 0), 2)  

            # Convert image back to PIL format  | หากต้องการดูรูปให้เปิด image_pil.show()  | COLOR_BGR2RGB , cv2.COLOR_BGR2RGB , cv2.COLOR_HSV2RGB
            # image_pil = Image.fromarray(cv2.cvtColor(image, cv2.COLOR_BGR2RGB))   #(กรณีทดสอบดูรูปแรกสุด)
            # image_pil.show()
            image_new_pil = Image.fromarray(image)
            # image_pil.show()

            # แปลงภาพ PIL เป็น base64
            image_old_pil = Image.fromarray(cv2.cvtColor(image_old, cv2.COLOR_BGR2RGB))
            buffered_old = io.BytesIO()
            image_old_pil.save(buffered_old, format="PNG")
            old_image_base64 = base64.b64encode(buffered_old.getvalue()).decode('utf-8')

            buffered_new = io.BytesIO()
            image_new_pil.save(buffered_new, format="PNG")
            new_image_base64 = base64.b64encode(buffered_new.getvalue()).decode('utf-8')

            #แสดงผลบน log.html 
            BuiltIn().log(f'''
            <table>
                <tr>
                    <td style="padding-right: 35px;"><h3>Old Image:</h3><img src="data:image/png;base64,{old_image_base64}" width="400px"></td>
                    <td style="padding-right: 35px; text-align: center; vertical-align: middle;"><h2 style="font-size: 125px; color: green;">&#11162;</h2></td>
                    <td><h3>New Image:</h3><img src="data:image/png;base64,{new_image_base64}" width="400px"></td>
                </tr>
            </table>
            ''', html=True)


            return  len(filtered_contours) , filtered_contours
        
        
        
    def highlight_contours_objects(self, image_path , contours, indices):
        """
        ***|    Description     |***
        |   *`Highlight Contours Objects`*   |   ส่ง path รูปภาพ , contours และ indices เพื่อไป highlight object ที่ต้องการตาม indices ที่กำหนด |
 
        ***|    Example     |***
        | *`Highlight Contours Objects`* | image_path = path | contours = data filtered_contours | indices  = [array or list contours]
        | *`Highlight Contours Objects`* | image_path = C/domain/log.png | contours = data filtered_contours | indices  = 0 

        ***|    Parameters     |***
        -  image_path
        -  contours (data filtered_contours ที่ได้จาก keyword : 'Detect And Mask Objects')
        -  indices (array or list ของตัวที่ต้องการ ภายใน data filtered_contours )
 
        *`ทดสอบได้เฉพาะบางหน้าเท่านั้น ค่าบางค่าอาจเอาออกมาไม่ได้เนื่องจากขนาดใหญ่เกินไป`*
        """
        #read image with cv2.imread
        image = cv2.imread(image_path)
        image_old = cv2.imread(image_path)
        indices = [int(indices)]
        # Draw specified contours on the image with different colors
        for index in indices:
            color = (0, 0, 255)
            result_image = cv2.drawContours(image, contours, index, color, 3)

        # Convert image back to PIL format
        image_new_pil = Image.fromarray(cv2.cvtColor(image, cv2.COLOR_BGR2RGB))
        # result_image.show()

        # แปลงภาพ PIL เป็น base64
        image_old_pil = Image.fromarray(cv2.cvtColor(image_old, cv2.COLOR_BGR2RGB))
        buffered_old = io.BytesIO()
        image_old_pil.save(buffered_old, format="PNG")
        old_image_base64 = base64.b64encode(buffered_old.getvalue()).decode('utf-8')

        buffered_new = io.BytesIO()
        image_new_pil.save(buffered_new, format="PNG")
        new_image_base64 = base64.b64encode(buffered_new.getvalue()).decode('utf-8')

        #แสดงผลบน log.html 
        BuiltIn().log(f'''
            <table>
                <tr>
                    <td style="padding-right: 35px;"><h3>Old Image:</h3><img src="data:image/png;base64,{old_image_base64}" width="400px"></td>
                    <td style="padding-right: 35px; text-align: center; vertical-align: middle;"><h2 style="font-size: 125px; color: green;">&#11162;</h2></td>
                    <td><h3>New Image:</h3><img src="data:image/png;base64,{new_image_base64}" width="400px"></td>
                </tr>
            </table>
        ''', html=True)

        return image_new_pil

    def get_average_bgr_color_within_contour(self, image_path , contour):
        """
        ***|    Description     |***
        |   *`Get Average Bgr Color Within Contour`*   |   ส่ง path รูปภาพ , contours เพื่อไปคำนวณค่าสีเฉลี่ยภายใน contour โดยใช้ mask ค่าที่ได้จะเป็น B , G , R , Alpha|
 
        ***|    Example     |***
        | *`Highlight Contours Objects`* | image_path = C/domain/log.png | contours = contour[0] (เลือกเฉพาะ array contour ที่ต้องการจะส่ง)

        ***|    Parameters     |***
        -  image_path
        -  contours (data filtered_contours ที่ได้จาก Keyword : 'Detect And Mask Objects')

        *`ทดสอบได้เฉพาะบางหน้าเท่านั้น ค่าบางค่าอาจเอาออกมาไม่ได้เนื่องจากขนาดใหญ่เกินไป`*
        """
        #read image with cv2.imread
        image = cv2.imread(image_path)
        image_old = cv2.imread(image_path)
        # สร้าง mask ที่มีขนาดเท่ากับภาพ โดยมีค่าเริ่มต้นเป็นสีดำ (ค่า 0)
        mask = np.zeros(image.shape[:2], dtype=np.uint8)

        # วาด contour บน mask ด้วยสีขาว
        draw = cv2.drawContours(mask, [contour], -1, (255), thickness=cv2.FILLED)

        # คำนวณค่าสีเฉลี่ยภายใน contour โดยใช้ mask
        mean_color = cv2.mean(image, mask=mask)

        # Convert image back to PIL format
        # result_image = Image.fromarray(cv2.cvtColor(draw, cv2.COLOR_BGR2RGB))
        image_new_pil = Image.fromarray(draw)
        # result_image.show()

        # แปลงภาพ PIL เป็น base64
        image_old_pil = Image.fromarray(cv2.cvtColor(image_old, cv2.COLOR_BGR2RGB))
        buffered_old = io.BytesIO()
        image_old_pil.save(buffered_old, format="PNG")
        old_image_base64 = base64.b64encode(buffered_old.getvalue()).decode('utf-8')

        buffered_new = io.BytesIO()
        image_new_pil.save(buffered_new, format="PNG")
        new_image_base64 = base64.b64encode(buffered_new.getvalue()).decode('utf-8')

        #แสดงผลบน log.html
        BuiltIn().log(f'''
            <table>
                <tr>
                    <td style="padding-right: 35px;"><h3>Old Image:</h3><img src="data:image/png;base64,{old_image_base64}" width="400px"></td>
                    <td style="padding-right: 35px; text-align: center; vertical-align: middle;"><h2 style="font-size: 125px; color: green;">&#11162;</h2></td>
                    <td><h3>New Image:</h3><img src="data:image/png;base64,{new_image_base64}" width="400px"></td>
                </tr>
            </table>
        ''', html=True)

        return mean_color
    
    def convert_bgr_2_rgb(self, average_bgra ):
        """
        ***|    Description     |***
        |   *`Convert Bgr 2 Rgb`*   |   ส่ง averages blue green red alpha เพื่อไปแปลงเป็น numpy array และแปลงจาก BGR เป็น RGB |
 
        ***|    Example     |***
        | *`Convert Bgr 2 Rgb`* | image_path = C/domain/log.png | contours = contour[0] (เลือกเฉพาะ array contour ที่ต้องการจะส่ง)

        ***|    Parameters     |***
        -  average_bgra (ค่า mean สี ที่ได้จาก Keyword : 'Get Average Bgr Color Within Contour')

        *`.....`*
        """
        #ลบค่า alpha ออกจาก list average_bgra
        bgr_values = tuple(list(average_bgra)[:-1])

        # แปลงเป็น numpy array และแปลงจาก BGR เป็น RGB
        #ทำการ reshape array ให้มีขนาด (1, 1, 3) ซึ่งหมายถึงมี 1 แถว, 1 คอลัมน์, และ 3 ช่องสี (BGR)
        average_color_bgr_array = np.array(bgr_values, dtype=np.float32).reshape((1, 1, 3))

        #ใช้ฟังก์ชัน cvtColor ของ OpenCV เพื่อแปลงค่าจากสี BGR เป็นสี RGB
        average_color_rgb_array = cv2.cvtColor(average_color_bgr_array, cv2.COLOR_BGR2RGB)

        #ทำการ flatten array ให้เป็นแถวเดียว (1-dimensional array)
        average_color_rgb = average_color_rgb_array.flatten()

        return  average_color_rgb
    
    def convert_rgb_2_hsv(self, Red , Green , Blue ):
        """
        ***|    Description     |***
        |   *`Convert RGB 2 HSV`*   |   ส่ง Red , Green , Blue เพื่อไปแปลงเป็น เป็น  HSV (โดยค่า rgb สามารถหาได้จากการ
        นำรูปภาพไปเข้าเว็บ rgb Image เพื่อหาสีที่เราต้องการ https://imagecolorpicker.com/th or https://www.ginifab.com/feeds/pms/color_picker_from_image.th.php)

        ***|    Example     |***
        | *`Convert RGB 2 HSV`* | Red = 10 | Green = 10 | Blue = 10

        ***|    Parameters     |***
        -  Red (สีแดง)
        -  Green (สีเขียว)
        -  Blue (สีฟ้า)

        *`.....`*
        """
        # สี RGB
        rgb_color = np.uint8([[[Red, Green, Blue]]])

        # แปลงจาก RGB เป็น HSV
        hsv_color = cv2.cvtColor(rgb_color, cv2.COLOR_RGB2HSV)

        #reshape(-1) จะทำการแปลง array ให้เป็น array หนึ่งมิติ
        hsv_array = hsv_color.reshape(-1)

        h = hsv_array[0]
        s = hsv_array[1]
        v = hsv_array[2]

        hsv_array = [int(h),int(s),int(v)]

        return  hsv_array

    def rgb_2_matplotlib_color_name_and_hex(self, rgb_tuple ):
        """
        ***|    Description     |***
        |   *`Rgb 2 Matplotlib Color Name And Hex`*   |   ส่ง averages red green blue เพื่อไปคำนวณหา ชื่อสี และ ค่า hex ของสีนั้น |
 
        ***|    Example     |***
        | *`Rgb 2 Matplotlib Color Name And Hex`* | rgb_tuple = data rgb  (Ex : [181.69142  240.91144   21.810137])

        ***|    Parameters     |***
        -  rgb_tuple (ค่า mean สี ที่ได้จาก Keyword : 'Get Average Bgr Color Within Contour' และแปลงเป็น rgb ด้วย Keyword : 'convert_bgr_2_rgb')

        *`.....`*
        """
        # Normalize the RGB values to be between 0 and 1 | ปรับค่า RGB ให้เป็นช่วง 0-1 โดยการหารค่าของแต่ละช่องสีด้วย 255 เพื่อทำให้ข้อมูลมีมาตรฐานเดียวกัน ซึ่งช่วยให้การคำนวณระยะทางระหว่างสี
        rgb_normalized = tuple([x / 255.0 for x in rgb_tuple])

        # Find the closest color name and hex |  เป็นการคำนวณระยะทางแบบ Euclidean ระหว่างสีที่ normalized และสี CSS4 แต่ละสี
        closest_name = None
        closest_hex = None
        min_distance = float('inf')
        for name, hex_value in mcolors.CSS4_COLORS.items():
            color_rgb = mcolors.hex2color(hex_value)
            # closest_hex = hex_value
            distance = sum((component1 - component2) ** 2 for component1, component2 in zip(rgb_normalized, color_rgb))
            if distance < min_distance:
                min_distance = distance
                closest_name = name
                closest_hex = hex_value

        #ส่งคืนชื่อสีและค่า hex ของสีที่ใกล้เคียงที่สุด
        return closest_name , closest_hex
    
    def replace_capture_path(self, path ):
        """
        ***|    Description     |***
        |   *`Replace Capture Path`*   |   ส่ง path ที่ได้จากการ capture screen เข้ามา replace \\ เป็น / |
 
        ***|    Example     |***
        | *`Replace Capture Path`* | path = C\domain\log.png 

        ***|    Parameters     |***
        -  path  (ค่า path ที่ได้จาก Keyword : 'Capture Screen ของ AppiumFlutter หรือ Appium')

        *`.....`*
        """
        newpath = path.replace("\\", "/")

        return  newpath
    
    def crop_image(self, left:int, top:int, right:int, bottom:int, image_path: str):
        imgold = Image.open(image_path)
        imgnew = Image.open(image_path)
        crop_img = imgnew.crop((left, top, right, bottom))
        # crop_img.show()
        buffered_old = io.BytesIO()
        imgold.save(buffered_old, format="PNG")
        old_image_base64 = base64.b64encode(buffered_old.getvalue()).decode('utf-8')

        buffered_new = io.BytesIO()
        crop_img.save(buffered_new, format="PNG")
        new_image_base64 = base64.b64encode(buffered_new.getvalue()).decode('utf-8')

        #แสดงผลบน log.html
        BuiltIn().log(f'''
            <table>
                <tr>
                    <td style="padding-right: 35px;"><h3>Image:</h3><img src="data:image/png;base64,{old_image_base64}" width="400px"></td>
                    <td style="padding-right: 35px; text-align: center; vertical-align: middle;"><h2 style="font-size: 125px; color: green;">&#61;</h2></td>
                    <td style="padding-left: 35px;"><h3>Crop Image:</h3></td>
                    <td style="text-align: center; vertical-align: middle;"><img src="data:image/png;base64,{new_image_base64}" width="400px"></td>
                </tr>
            </table>
        ''', html=True)

        return crop_img
    
    def compare_images(self , crop_img , path_img_compare):
        """
        ***|    Description     |***
        |   *`Compare Images`*   |   ส่ง path รูปภาพ crop_img และ path_img_compare,
        เพื่อไปคำนวณค่าความแตกต่างของภาพ โดยหาก np.array ในตั้ง 2 ภาพเท่ากัน จะ pass
 
        ***|    Example     |***
        | *`Compare Images`* | crop_img  | path_img_compare | fail_threshold (default= 5)

        ***|    Parameters     |***
        -  crop_img (img ที่ได้จากการ crop แล้วจาก  def crop_image (type file PIL))
        -  path_img_compare  (path ที่เก็บรูปที่ไว้ตรวจสอบ)

        *`หากทดสอบใน ต่างรุ่นโทรศัพจะ Fail`*
        """
        np_image_crop = np.array(crop_img)
        compare_img = Image.open(path_img_compare)
        np_image_compare = np.array(compare_img)

        #แปลง PIL เป็น base 64 
        buffered_old = io.BytesIO()
        crop_img.save(buffered_old, format="PNG")
        old_image_base64 = base64.b64encode(buffered_old.getvalue()).decode('utf-8')

        buffered_new = io.BytesIO()
        compare_img.save(buffered_new, format="PNG")
        new_image_base64 = base64.b64encode(buffered_new.getvalue()).decode('utf-8')

        # Check if shapes of the two images are same
        if np_image_crop.shape == np_image_compare.shape:
            # Compare pixel by pixel
            if np.array_equal(np_image_crop, np_image_compare):
                BuiltIn().log('The images are exactly the same.')
                BuiltIn().log(f'''
                    <table>
                        <tr>
                            <td style="padding-right: 35px;"><h3>Actual Image:</h3><img src="data:image/png;base64,{old_image_base64}" width="400px"></td>
                            <td style="padding-right: 35px; text-align: center; vertical-align: middle;"><h2 style="font-size: 125px; color: green;">&#61;</h2></td>
                            <td><h3>Expect Image:</h3><img src="data:image/png;base64,{new_image_base64}" width="400px"></td>
                        </tr>
                    </table>
                ''', html=True)
            else:
                BuiltIn().log(f'''
                    <table>
                        <tr>
                            <td style="padding-right: 35px;"><h3>Actual Image:</h3><img src="data:image/png;base64,{old_image_base64}" width="400px"></td>
                            <td style="padding-right: 35px; text-align: center; vertical-align: middle;"><h2 style="font-size: 125px; color: green;">&#8800;</h2></td>
                            <td><h3>Expect Image:</h3><img src="data:image/png;base64,{new_image_base64}" width="400px"></td>
                        </tr>
                    </table>
                ''', html=True)
                BuiltIn().run_keyword('Fail','The images are different.')
        else:
            BuiltIn().log(f'''
                <table>
                    <tr>
                        <td style="padding-right: 35px;"><h3>Actual Image:</h3><img src="data:image/png;base64,{old_image_base64}" width="400px"></td>
                        <td style="padding-right: 35px; text-align: center; vertical-align: middle;"><h2 style="font-size: 125px; color: green;">&#8800;</h2></td>
                        <td style="padding-left: 35px;"><h3>Expect Image:</h3></td>
                        <td style="text-align: center; vertical-align: middle;"><img src="data:image/png;base64,{new_image_base64}" width="400px"></td>
                    </tr>
                </table>
            ''', html=True)
            BuiltIn().run_keyword('Fail','The images have different sizes and cannot be compared directly.')

    def find_fullpath_with_filename(self,filename:str,search_path:str):
        for root, dirs, files in os.walk(search_path):
            if filename in files:
                return os.path.join(root, filename)
        return None

    def save_image_on_directory(self,image,saveonpath):
        #Check isinstance PIL
        if self._is_pil_image(image):
            BuiltIn().log('This is a PIL Image object.')
            BuiltIn().log(f'Size: {image.size}')
            BuiltIn().log(f'Format: {image.format}')
            BuiltIn().log(f'Mode: {image.mode}')
            image.save(saveonpath)
            BuiltIn().log(f'Success Save Image On Path : {saveonpath}')
        else:
            BuiltIn().run_keyword('Fail','This is not a PIL Image object.')

    def compare_images_percent(self , path_img_actual:str , path_img_compare:str ,fail_threshold:int=5):
        """
        ***|    Description     |***
        |   *`Compare Images Percent`*   |   ส่ง path รูปภาพ actual และ compare, fail_threshold ค่าที่ยอมรับ 
        เพื่อไปคำนวณค่าความแตกต่างของภาพ หากค่าที่ได้มากกว่า fail_threshold ที่ให้ไปจะ fail
 
        ***|    Example     |***
        | *`Compare Images Percent`* | path_img_actual  | path_img_compare | fail_threshold (default= 5)

        ***|    Parameters     |***
        -  path_img_compare (path ที่เก็บรูป compare)
        -  path_img_actual  (path ที่เก็บรูปที่จะทดสอบ)
        -  fail_threshold (ค่าที่ยอบรับความต่างภาพ)

        ***|    Doc     |***
        -  difference (ตัวแปรนี้เก็บค่าความแตกต่างของภาพจากฟังก์ชัน cv2.absdiff() ซึ่งคำนวณความแตกต่างแบบแอบโซลูตระหว่างสองภาพ.)
        -  gray_difference (ตัวแปรนี้เก็บภาพความแตกต่างที่ถูกแปลงเป็นภาพสีเทา (cv2.cvtColor()) เพื่อลดความซับซ้อนในการประมวลผลต่อไป.)
        -  threshold_value, thresh  (threshold_value คือค่าเกณฑ์ที่ใช้ในการแยกพิกเซลที่แตกต่างออกจากภาพความแตกต่าง. thresh คือภาพที่ได้จากการใช้เกณฑ์นี้ซึ่งแสดงพิกเซลที่แตกต่างชัดเจน.)
        -  num_diff_pixels (ตัวแปรนี้เก็บจำนวนพิกเซลที่ถูกตัดสินว่าแตกต่างกันอย่างชัดเจนหลังจากใช้เกณฑ์.)
        -  total_pixels, percentage_difference (total_pixels คือจำนวนพิกเซลทั้งหมดใน thresh. percentage_difference คือเปอร์เซ็นต์ของพิกเซลที่แตกต่างกันเมื่อเทียบกับจำนวนพิกเซลทั้งหมด.)
        -  fail_threshold   (เกณฑ์ที่ใช้ตัดสินว่าภาพแตกต่างกันมากเพียงใดจึงจะถือว่าการเปรียบเทียบล้มเหลว.)
        
        *`หากทดสอบใน ต่างรุ่นโทรศัพ อาจเกิด ค่าความแตกต่างของภาพที่มาก`*
        """
        #ใช้สำหรับประมวลตวามต่าง
        image_actual = cv2.imread(path_img_actual)
        image_compare = cv2.imread(path_img_compare)
        

        #ใช้แสดงรูป
        actual_img = Image.open(path_img_actual)
        compare_img = Image.open(path_img_compare)
        #แปลง PIL เป็น base 64 
        buffered_old = io.BytesIO()
        actual_img.save(buffered_old, format="PNG")
        old_image_base64 = base64.b64encode(buffered_old.getvalue()).decode('utf-8')

        buffered_new = io.BytesIO()
        image_compare.save(buffered_new, format="PNG")
        new_image_base64 = base64.b64encode(buffered_new.getvalue()).decode('utf-8')


        # ตรวจสอบขนาดภาพ และปรับให้เท่ากัน
        if image_actual.shape != image_compare.shape:
            # ปรับขนาด image2 ให้เท่ากับ image1
            image_compare = cv2.resize(image_compare, (image_actual.shape[1], image_actual.shape[0]))

        # คำนวณความแตกต่างของภาพ
        difference = cv2.absdiff(image_actual, image_compare)

        # แปลงภาพความแตกต่างเป็นภาพสีเทา
        gray_difference = cv2.cvtColor(difference, cv2.COLOR_BGR2GRAY)

        # ตั้งค่าความต่างที่ถือว่ายอมรับได้ (เกณฑ์)
        threshold_value = 30  # กำหนดเกณฑ์ความต่างที่ยอมรับได้

        # ทำ threshold เพื่อแยกพิกเซลที่แตกต่างกันอย่างเห็นได้ชัด
        _, thresh = cv2.threshold(gray_difference, threshold_value, 255, cv2.THRESH_BINARY)

        # คำนวณจำนวนพิกเซลที่แตกต่างกัน
        num_diff_pixels = np.count_nonzero(thresh)

        # คำนวณเปอร์เซ็นต์ของพิกเซลที่แตกต่างกัน
        total_pixels = thresh.size
        percentage_difference = (num_diff_pixels / total_pixels) * 100

        fail_thresholdFix = fail_threshold  # เกณฑ์ความแตกต่างที่ยอมรับได้ 5% (defual = 5 %)
        if percentage_difference > fail_thresholdFix:
            BuiltIn().log('Test fail: The image difference is too high.')
            BuiltIn().log(f'''
                    <table>
                        <tr>
                            <td style="padding-right: 35px;"><h3>Actual Image:</h3><img src="data:image/png;base64,{old_image_base64}" width="400px"></td>
                            <td style="padding-right: 35px; text-align: center; vertical-align: middle;"><h2 style="font-size: 125px; color: green;">&#8800;</h2></td>
                            <td><h3>Expect Image:</h3><img src="data:image/png;base64,{new_image_base64}" width="400px"></td>
                        </tr>
                    </table>
            ''', html=True)
            BuiltIn().run_keyword('Fail',f'Percentage difference: {percentage_difference:.2f}% More Than Fail threshold : {fail_thresholdFix:.2f}')
        else:
            BuiltIn().log('Test pass: The images are considered similar.')
            BuiltIn().log(f"Percentage difference: {percentage_difference:.2f}% Less Than Fail threshold : {fail_thresholdFix:.2f}")
            BuiltIn().log(f'''
                    <table>
                        <tr>
                            <td style="padding-right: 35px;"><h3>Actual Image:</h3><img src="data:image/png;base64,{old_image_base64}" width="400px"></td>
                            <td style="padding-right: 35px; text-align: center; vertical-align: middle;"><h2 style="font-size: 125px; color: green;">&#61;</h2></td>
                            <td><h3>Expect Image:</h3><img src="data:image/png;base64,{new_image_base64}" width="400px"></td>
                        </tr>
                    </table>
            ''', html=True)

    #PRIVATE_FUNCTION
    
    def _is_pil_image(self,obj):
            return isinstance(obj, Image.Image)
        

    


# -*- coding: utf-8 -*-
from .connectionmanagement import ConnectionManagement

cache_app = ConnectionManagement()

class KeyEvent():

    def __init__(self):
        #เนื่องจากปัญหาเรื่องโครงสร้าง structure เลยยังไม่สามารถใช้ได้
        pass

    def native_press_keycode(self, keycode, metastate=None):
        """Sends a press of keycode to the device.

        Android only.

        Possible keycodes & meta states can be found in
        http://developer.android.com/reference/android/view/KeyEvent.html

        Meta state describe the pressed state of key modifiers such as
        Shift, Ctrl & Alt keys. The Meta State is an integer in which each
        bit set to 1 represents a pressed meta key.

        For example
        - META_SHIFT_ON = 1
        - META_ALT_ON = 2

        | metastate=1 --> Shift is pressed
        | metastate=2 --> Alt is pressed
        | metastate=3 --> Shift+Alt is pressed

         - _keycode- - the keycode to be sent to the device
         - _metastate- - status of the meta keys

        ============================================================

        ฟังก์ชันที่ใช้สำหรับการส่งการกดปุ่มคีย์โค้ดไปยังอุปกรณ์ Android ในการทดสอบอัตโนมัติด้วยเฟรมเวิร์ก Appium 
        ซึ่งเป็นเฟรมเวิร์กยอดนิยมสำหรับการทดสอบแอปพลิเคชันบนมือถือ

        - keycode: รหัสของปุ่มที่ต้องการกด รหัสเหล่านี้สามารถหาได้จากเอกสารของ Android ที่
          http://developer.android.com/reference/android/view/KeyEvent.html

        - metastate: สถานะของปุ่มตัวช่วยเช่น Shift, Ctrl, หรือ Alt 
        ซึ่งถูกแสดงด้วยตัวเลขที่แต่ละบิตที่ถูกตั้งค่าเป็น 1 แทนสถานะที่กดของปุ่มตัวช่วยนั้น ๆ
        """
        driver = cache_app._current_application()
        driver.press_keycode(keycode, metastate)


# -*- coding: utf-8 -*-

import os
import robot

from .connectionmanagement import ConnectionManagement
from AppiumLibrary.keywords._logging import _LoggingKeywords

cache_app = ConnectionManagement()
log = _LoggingKeywords()

class CaptureScreenShot:

    def __init__(self):
        pass
    
    #KeyWord

    def native_capture_page_screenshot(self, filename=None):
        """Takes a screenshot of the current page and embeds it into the log.

        `filename` argument specifies the name of the file to write the
        screenshot into. If no `filename` is given, the screenshot will be
        embedded as Base64 image to the log.html. In this case no file is created in the filesystem.

        Warning: this behavior is new in 1.7. Previously if no filename was given
        the screenshots where stored as separate files named `appium-screenshot-<counter>.png`

        Example:

        |t_capture_page_screenshot | None |

        =========================================================

        ถ่ายภาพหน้าจอของหน้าปัจจุบันและฝังลงในบันทึก

        อาร์กิวเมนต์ filename ระบุชื่อของไฟล์ที่จะเขียนภาพหน้าจอลงไป หากไม่มีการระบุ filename, ภาพหน้าจอจะถูกฝังเป็นรูปภาพ Base64 ลงใน log.html
        ในกรณีนี้ไม่มีไฟล์ใดถูกสร้างขึ้นในระบบไฟล์

        คำเตือน: พฤติกรรมนี้เป็นใหม่ในเวอร์ชัน 1.7 ก่อนหน้านี้หากไม่มีการระบุชื่อไฟล์ ภาพหน้าจอจะถูกเก็บเป็นไฟล์แยกต่างหากชื่อว่า 
        appium-screenshot-<counter>.png

        ตัวอย่างการใช้งาน:

        |t_capture_page_screenshot | None |


        """
        if filename:
            path, link = self._get_screenshot_paths(filename)

            if hasattr(cache_app._current_application(), 'get_screenshot_as_file'):
                cache_app._current_application().get_screenshot_as_file(path)
            else:
                cache_app._current_application().save_screenshot(path)

            # Image is shown on its own row and thus prev row is closed on purpose
            log._html('</td></tr><tr><td colspan="3"><a href="%s">'
                       '<img src="%s" width="800px"></a>' % (link, link))
            return path
        else:
            base64_screenshot = cache_app._current_application().get_screenshot_as_base64()
            log._html('</td></tr><tr><td colspan="3">'
                       '<img src="data:image/png;base64, %s" width="800px">' % base64_screenshot)
            return None

    #Private_Function
        
    def _get_screenshot_paths(self, filename):
        filename = filename.replace('/', os.sep)
        logdir = log._get_log_dir()
        path = os.path.join(logdir, filename)
        link = robot.utils.get_link_path(path, logdir)
        return path, link



# -*- coding: utf-8 -*-
import bson
import logging
import pytz
import re
import base64
import json

from datetime import datetime

class ConvertObject:

    def __init__(self):
        #เนื่องจากปัญหาเรื่องโครงสร้าง structure เลยยังไม่สามารถใช้ได้
        # self._bi = BuiltIn()
        pass

    #KeyWord

    def convert_bson_to_json(self, bson_data):
        """ Owner : tassana.k@epic-consulting.net
        Convert BSON data to a Python object (dictionary or list) 
        without converting it to a JSON string.

        ==========================================================

        แปลงข้อมูล BSON เป็นอ็อบเจกต์ Python (dictionary หรือ list) 
        โดยไม่ต้องแปลงเป็นสตริง JSON
        """
        try:
            if isinstance(bson_data, bytes):
                bson_data = bson.loads(bson_data)
            # Convert BSON directly to a Python object without converting to a JSON string
            return bson_data  # This returns a Python dictionary or list of dictionaries
        except Exception as e:
            return f"Failed to convert BSON to JSON object: {str(e)}"

    def log_tree_structure_data(self, data):
        """ Owner : tassana.k@epic-consulting.net
        ***|    Description     |***
        |   *`Log Tree Data`*   |   เป็น Keyword สำหรับ Log Data ให้ออกมาเป็น Tree Data |

        ***|    Example     |***
        | *`Log Tree Data`* | *`${value}`* |
        
        ***|    Parameters     |***
            - **`data`**  ข้อมูล Json.
        """
        self._tree_structure(data)

    def convert_time_to_local_timezone(self,date,timezone='Asia/Bangkok'):
        """ Owner : suthasinee.k@arcadiaapm.com
        ***|    Description     |***
        |   มีไว้สำหรับรับเวลาที่ได้มาแล้วแปลงเป็นเวลาของ TimeZone 
        |   โดย Default คือ Asia/Bangkok
        """
        # กำหนด timezone เป็น UTC
        utc_timezone = pytz.timezone('UTC')
        # แปลง timezone จาก UTC เป็น local timezone
        local_timezone = pytz.timezone(timezone)  
        # ตรวจสอบว่าข้อมูลนำเข้าเป็น datetime object หรือไม่
        if isinstance(date, datetime):
            date_time = date
            local_date_time = utc_timezone.localize(date).astimezone(local_timezone)
        else:
            # ตรวจสอบว่าข้อมูลนำเข้าเป็นรูปแบบที่ถูกต้องหรือไม่
            if not self._is_valid_iso_format(date):
                raise ValueError("Invalid date format. Please provide the date in the format 'YYYY-MM-DDTHH:MM:SS.sss+HH:MM'")
        
            # แปลง string เป็น datetime object
            date_time = datetime.fromisoformat(date)
            local_date_time = date_time.astimezone(local_timezone)
        
        return local_date_time
    
    #(ยังใช้ไม่ได้อยู่ในช่วงทดสอบ)
    def convert_get_render_tree_to_json(self,tree):
        """ Owner : pornpawit.s@chareontut.co.th
        ***|    Description     |***
        |   มีไว้สำหรับนำข้อมูล get render tree แปลง
        |   เป็น json (ยังใช้ไม่ได้อยู่ในช่วงทดสอบ)
        """
        # แปลงข้อมูล Render Tree เป็น JSON
        render_tree_json = json.dumps(tree, indent=4)
        with open('render_tree.json', 'w') as json_file:
            json_file.write(render_tree_json)
        
        with open('render_tree.json', 'r') as json_file:
            data = json.load(json_file)
            print(data)
        


    #Private Function

    def _tree_structure(self, data, level=0, prefix=''):
        indent = "│   " * level
        branch = "├── " if level > 0 else ""
        if isinstance(data, dict):
            for key, value in data.items():
                if isinstance(value, (dict, list)):
                    logging.info(f"{indent}{branch}{key}")
                    self._tree_structure(value, level + 1)
                else:
                    logging.info(f"{indent}{branch}{key}: {value}")
        elif isinstance(data, list):
            for i, item in enumerate(data):
                logging.info(f"{indent}├── [{i}]")
                self._tree_structure(item, level + 1)
        else:
            logging.info(f"{indent}{branch}{prefix}{data}")

    def _is_valid_iso_format(self,date_str):
        """ Owner : suthasinee.k@arcadiaapm.com
        ***|    Description     |***
        | 
        """
        pattern = re.compile(r'^\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}\.\d{3}\+\d{2}:\d{2}$')
        return bool(pattern.match(date_str))


# -*- coding: utf-8 -*-

import robot
import inspect
import logging
import json
import requests

from appium import webdriver
from robot.libraries.BuiltIn import BuiltIn
from TineAutomationToolkit.utils.applicationcache import ApplicationCache
from AppiumLibrary.keywords._logging import _LoggingKeywords

cache_app = BuiltIn()
method_cache_app = ApplicationCache()
log = _LoggingKeywords()

class ConnectionManagement:

    def __init__(self):
        #เนื่องจากปัญหาเรื่องโครงสร้าง structure เลยยังไม่สามารถใช้ได้
        # self._bi = BuiltIn()
        pass
    #KeyWord
    
    def native_close_application_session(self,alias=None):
        """Close Application And Quit Seesion

        =========================================================
        
        ปิดแอพปัจจุบันและปิดเซสชัน"""
        driver = self._current_application()
        log._debug('Closing application with session id %s' % self._current_application().session_id)
        method_cache_app.register(driver,alias)
        method_cache_app.close_all()

    def native_background_application(self, seconds=5):
        """
        Puts the application in the background on the device for a certain
        duration.

        =========================================================

        วางแอปพลิเคชันไว้ในพื้นหลังของอุปกรณ์เป็นระยะเวลาหนึ่ง.
        """
        self._current_application().background_app(seconds)
        
    def native_activate_application(self, app_id):
        """
        Activates the application if it is not running or is running in the background.
        Args:
         - app_id - BundleId for iOS. Package name for Android.

        New in AppiumLibrary v2

        =========================================================

        
        เปิดใช้งานแอปพลิเคชันหากมันไม่ได้รันอยู่หรือรันอยู่ในพื้นหลัง
        อาร์กิวเมนต์:
        - app_id - BundleId สำหรับ iOS, ชื่อแพ็คเกจสำหรับ Android.
        """
        self._current_application().activate_app(app_id)

    def native_terminate_application(self, app_id):
        """
        Terminate the given app on the device

        Args:
         - app_id - BundleId for iOS. Package name for Android.

        New in AppiumLibrary v2

        =========================================================
        
        ยุติแอปพลิเคชันที่กำหนดไว้บนอุปกรณ์แต่ยัง active อยู่ในระบบ
        สามารถเปิดต่อได้

        อาร์กิวเมนต์:  

        app_id - BundleId สำหรับ iOS, ชื่อแพ็คเกจสำหรับ Android.
        ใหม่ใน AppiumLibrary เวอร์ชัน 2
        """
        return self._current_application().terminate_app(app_id)
    
    def native_get_window_height(self):
        """Get current device height.

        Example:
        | ${width}       | Get Window Width |
        | ${height}      | Get Window Height |
        | Click A Point  | ${width}         | ${height} |

        New in AppiumLibrary 1.4.5
        """
        return self._current_application().get_window_size()['height']
    
    def native_get_window_width(self):
        """Get current device width.

        Example:
        | ${width}       | Get Window Width |
        | ${height}      | Get Window Height |
        | Click A Point  | ${width}          | ${height} |

        New in AppiumLibrary 1.4.5
        """
        return self._current_application().get_window_size()['width']
    
    def commond_install_app(self, app_path, app_package):
        """ *******Not available wait for update flutter*******
        Install App via Appium
        
        Android .

        - app_path - path to app (.apk)
        - app_package - package of install app to verify

        Ios .

        - app_path - path to app (.app | .ipa)
        - bundleId - package of install app to verify
        """
        driver = self._current_application()
        driver.install_app(app_path)
        return driver.is_app_installed(app_package)
    
    def get_source(self):
        """Returns the entire source of the current page.
        
        =========================================================

        ฟังก์ชันนี้จะส่งคืนสตริงที่มีซอร์สโค้ด HTML ของหน้าเว็บที่กำลังแสดงอยู่ในเบราว์เซอร์ในขณะนั้น 
        ซึ่งสามารถนำไปใช้ในการตรวจสอบหรือวิเคราะห์โครงสร้างหรือเนื้อหาของหน้าเว็บได้
        """
        return self._current_application().page_source
    
    def log_source(self, loglevel='INFO'):
        """Logs and returns the entire html source of the current page or frame.

        The `loglevel` argument defines the used log level. Valid log levels are
        `WARN`, `INFO` (default), `DEBUG`, `TRACE` and `NONE` (no logging).
        """
        ll = loglevel.upper()
        if ll == 'NONE':
            return ''
        else:
            if  "run_keyword_and_ignore_error" not in [check_error_ignored[3] for check_error_ignored in inspect.stack()]:
                source = self._current_application().page_source
                log._log(source, ll)
                return source
            else:
                return ''
            
    def get_driver(self):
        """*******Not available wait for update flutter*******
        Connect Session(Don't Create New Session & )
        """
        current_app_caps = self._current_application()
        driver = webdriver.Remote('http://localhost:4723/wd/hub', desired_capabilities=current_app_caps, direct_connection=True)
        return driver
    

    def execute_script(self, script, **kwargs):
        """
        Execute a variety of native, mobile commands that aren't associated
        with a specific endpoint. See [https://appium.io/docs/en/commands/mobile-command/|Appium Mobile Command]
        for more details.

        Example:
        | &{scrollGesture}  |  create dictionary  |  left=${50}  |  top=${150}  |  width=${50}  |  height=${200}  |  direction=down  |  percent=${100}  |
        | Sleep             |  1                  |
        | Execute Script    |  mobile: scrollGesture  |  &{scrollGesture}  |

        Updated in AppiumLibrary 2
        """
        if kwargs:
            log._info(f"Provided dictionary: {kwargs}")

        return self._current_application().execute_script(script, kwargs)
    
    def get_appium_session_id(self):
        """Returns the current session ID as a reference"""
        log._info("Appium Session ID: " + self._current_application().session_id)
        return self._current_application().session_id
    
    def send_session_status_to_browserstack(self):
        """
        ***|    Description     |***
        |   *`Send Session Status To BrowserStack`*   |   ส่งสถานะและข้อความไปยัง BrowserStack ตาม session_id, status, และ message ที่กำหนด |
 
        ***|    Example     |***
        | *`Send Session Status To BrowserStack`* |
 
        ***|    Parameters     |***
        - ไม่มีพารามิเตอร์
 
        *`Create By Tassana Khrueawan (tassana.khr@gmail.com | Tel:097-170-3730)`*
        """
        # ประกาศค่า variables ที่ใช้ในการเชื่อมต่อกับ BrowserStack
        BROWSERSTACK_USERNAME = 'juralakp_p2Vx0U'
        BROWSERSTACK_ACCESS_KEY = 'PPed2Tj3ZNZqaGw6kLyw'
        BROWSERSTACK_URL = f"https://{BROWSERSTACK_USERNAME}:{BROWSERSTACK_ACCESS_KEY}@hub-cloud.browserstack.com/wd/hub"

        session_id = BuiltIn().get_variable_value("${SESSION_ID}")
        # session_id = self.get_appium_session_id()
        
        # ดึงค่า TEST STATUS และ TEST MESSAGE จาก Robot Framework
        test_status = BuiltIn().get_variable_value("${TEST STATUS}")
        test_message = BuiltIn().get_variable_value("${TEST MESSAGE}")

        # ถ้า TEST STATUS หรือ TEST MESSAGE เป็น None ให้ดึงข้อมูลจาก get_variables
        if not test_status or not test_message:
            variables = BuiltIn().get_variables()
            test_status = variables.get("${TEST STATUS}", "FAIL")
            test_message = variables.get("${TEST MESSAGE}", "")
 
        # ตรวจสอบค่า test_status และ test_message
        logging.info(f"Test Status: {test_status}")
        logging.info(f"Test Message: {test_message}")
 
        # กำหนดสถานะการทดสอบเป็น "passed" ถ้าผลลัพธ์คือ "PASS" มิฉะนั้นจะเป็น "failed"
        status = "passed" if test_status == "PASS" else "failed"
 
        # กำหนดข้อความเป็น "PASS" ถ้าสถานะคือ "passed" มิฉะนั้นจะใช้ข้อความจาก test_message
        message = "PASS" if status == "passed" else test_message
 
        # Log the status and message for debugging
        logging.info(f"Sending to BrowserStack: status={status}, message={message}")
 
        # สร้าง URL สำหรับการส่งคำขอไปยัง BrowserStack
        url = f"{BROWSERSTACK_URL}/session/{session_id}/execute"
 
        # สร้าง payload ในรูปแบบ JSON ที่จะส่งไปยัง BrowserStack
        payload = json.dumps({
            "script": f'browserstack_executor: {{"action": "setSessionStatus", "arguments": {{"status": "{status}", "reason": "{message}"}}}}',
            "args": []
        })
 
        # ตั้งค่า headers สำหรับคำขอ HTTP POST
        headers = {
            'Content-Type': 'application/json'
        }
 
        # ส่งคำขอ HTTP POST ไปยัง BrowserStack
        response = requests.post(url, headers=headers, data=payload)
 
        # ตรวจสอบสถานะการตอบกลับและบันทึกข้อมูล
        if response.status_code == 200:
            logging.info(f"Successfully sent status to BrowserStack: {status}, {message}")
        else:
            logging.error(f"Failed to send status to BrowserStack: {response.status_code}, {response.content}")

    #PRIVATE_FUNCTION
        
    def _current_application(self):
        """Return the instance of the current application
        From AppiumFlutterLibrary

        =========================================================

        คืนค่าอินสแตนซ์ของแอปพลิเคชันปัจจุบัน
        จาก AppiumFlutterLibrary
        """
        return cache_app.get_library_instance('AppiumFlutterLibrary')._current_application()
        # return self._bi.get_library_instance('AppiumFlutterLibrary')._current_application()

    def _get_platform(self):
        try:
            platform_name = self._current_application().desired_capabilities['platformName']
        except Exception as e:
            raise e
        return platform_name.lower()


        